#!/usr/bin/env python

#=====================================================================================
#
# Description: This script provides a Fourier transformation of a density matrix block
#              from ReSpect output at a certain specified freqency.
#              The output is a readable file with the transition matrix.
#
# Author:      Marius Kadek (marius.kadek@uit.no)
#              Lukas Konecny (lukas.konecny1@gmail.com)
#
# Revision:    1.0 (07/2014)
#
#=====================================================================================

import os
import sys
import math
import cmath
import numpy as np
import matplotlib.pyplot as plt
from optparse import OptionParser

class InputParameters:
  def parse_arguments( self ):
    usage = './analysis_matrix_block.py --files=/path/to/the/input/file_x'
    parser = OptionParser( usage )
    parser.add_option( '-f', '--files',
                       action='store',
                       type='string',
                       dest='path',
                       default=None,
                       help='Absolute path to the input file without "_x.out" extension. [default: %default]',
                       metavar='STRING' )
    parser.add_option( '-w', '--freq',
                       action='store',
                       type='float',
                       dest='frequency',
                       default=None,
                       help='Frequency of the transition (in Hartrees) in spectrum to be analysed. [default: %default]',
                       metavar='FLOAT' )
    parser.add_option( '-d', '--damping',
                       action='store',
                       type='float',
                       dest='dampingConstant',
                       default=0.004,
                       help='"gamma" constant (in a.u.) used to artificially damp the time signal, ' +
                            'i.e. multiply it by exp(-gamma*time). [default: %default]',
                       metavar='FLOAT' )
    (options, args) = parser.parse_args()
    self.path            = options.path
    self.frequency       = options.frequency
    self.dampingConstant = options.dampingConstant
    if self.path == None:
      print ' ERROR: no input file specified.'
      print '        ( for more info use "--help" )'
      sys.exit()
    if self.frequency == None:
      print ' ERROR: no transition frequency specified.'
      print '        ( for more info use "--help" )'
      sys.exit()

def get_files( path ):
  files = {}
  if os.path.isfile( path + '_x.out' ):
    files['x'] = path + '_x.out'
    print ' - file %s found' % files['x']
  if os.path.isfile( path + '_y.out' ):
    files['y'] = path + '_y.out'
    print ' - file %s found' % files['y']
  if os.path.isfile( path + '_z.out' ):
    files['z'] = path + '_z.out'
    print ' - file %s found' % files['z']
  if len( files ) == 0:
    print ' ERROR: File %s_x.out does not exist' % path
    print ' ERROR: File %s_y.out does not exist' % path
    print ' ERROR: File %s_z.out does not exist' % path
    print '        ( run "./analysis_matrix_block.py --help" to learn more )'
    sys.exit()
  return files

def get_symmetry( files ):
  if len( files ) == 1:
    print ' => assuming spherical symmetry'
    symmetry = 's'
  elif len( files ) == 2:
    print ' => assuming cylindrical symmetry'
    coordinates = files.keys()
    info = ' Select the symmetry axis of the molecule (%s or %s): ' % (coordinates[0], coordinates[1])
    symmetryAxis = raw_input( info ).lower()
    if symmetryAxis == coordinates[0] or symmetryAxis == coordinates[1]:
      symmetry = 'c' + symmetryAxis
    else:
      print ' ERROR: Wrong input specified'
      sys.exit()
  else:
    symmetry = 'n'
  return symmetry

class MatrixBlock: 
  def initialize_matrix_block( self, path ):
    with open( path ) as f:
      for row in f:
        row = row.strip( '\n' )
        if 'first row to print:' in row:
          row = row.split()
          self.rowStart = int( row[-1] )
          print ' - block rows from: %i' % self.rowStart
        if 'last  row to print:' in row:
          row = row.split()
          self.rowStop  = int( row[-1] )
          print ' - block rows to: %i' % self.rowStop
        if 'first column to print:' in row:
          row = row.split()
          self.colStart = int( row[-1] )
          print ' - block columns from: %i' % self.colStart
        if 'last  column to print:' in row:
          row = row.split()
          self.colStop  = int( row[-1] )
          print ' - block columns to: %i' % self.colStop
        if 'Total number of spherical GTO functions' in row:
          row = row.split()
          self.matrixDimension = int( row[-1] )
        if 'time step length:' in row:
          row = row.split()
          self.timeStep = float( row[-1] )
        if 'Step:' in row:
          break
    numberElements = ( self.colStop - self.colStart + 1 ) * ( self.rowStop - self.rowStart + 1 )
    self.block = np.matrix([complex(0.0,0.0)] * numberElements, dtype=np.complex_)
    self.block.shape = ( (self.rowStop-self.rowStart+1), (self.colStop-self.colStart+1) )
  def transform_matrix_block( self, path, frequency, damping, symFactor ):
    with open( path ) as f:
      for row in f:
        row = row.strip( '\n' )
        if 'Step:' in row:
          row = row.split()
          time = float( row[2] )
          expFactor = cmath.exp( complex( -damping*time, frequency*time) ) * symFactor
        if 'Element:' in row:
          row = row.split()
          valDmu = float( row[-1] )
          element = valDmu
          element *= expFactor
          self.block[int(row[1])-self.rowStart,int(row[2])-self.colStart] += element
  def get_trace( self, damping ):
    self.trace = 0.0 + 0.0*1j
    for i in range((self.rowStop-self.rowStart+1)):
      for j in range((self.colStop-self.colStart+1)):
        self.trace += self.block[i,j]
    self.trace /= 3.0
    self.trace *= self.timeStep*damping
  def print_block_to_file( self, path, frequency, damping ):
    with open(path, 'w' ) as output:
      output.write( 'time step length [a.u.]                 = %e\n' % self.timeStep   )
      output.write( 'analysed transition frequency [Hartree] = %e\n' % frequency       )
      output.write( 'damping constant (gamma)      [a.u.]    = %e\n' % damping         )
      output.write( 'Peak intensity from analyzed block      = %e\n' % self.trace.imag )
      output.write( 'MO matrix dimension %i \n' % self.matrixDimension )
      output.write( 'TmatBlock \n' )
      for i in range((self.rowStop-self.rowStart+1)):
        for j in range((self.colStop-self.colStart+1)):
          row = i + self.rowStart
          col = j + self.colStart
#         Real part is set to 0 to avoid mistakes, the abosrption spectra is in the imaginary part
          output.write( '%5i %5i %18.12f %18.12f\n' % ( row, col, 0.0, self.block[i,j].imag ) )
    print '- file %s written successfully' % path

#==================================================
#                       MAIN
#==================================================

inputParameters = InputParameters()
inputParameters.parse_arguments()

inputFiles = get_files( inputParameters.path )
symmetry   = get_symmetry( inputFiles )

matrixBlock = MatrixBlock()

matrixBlock.initialize_matrix_block( inputFiles[inputFiles.keys()[0]] )

for pulseDirection in inputFiles:
  symFactor = 1.0
  if symmetry[0] == 's':
    symFactor = 3.0
  if symmetry[0] == 'c':
    if symmetry[1] != pulseDirection:
      symFactor = 2.0
  matrixBlock.transform_matrix_block( inputFiles[pulseDirection], inputParameters.frequency, inputParameters.dampingConstant, symFactor )

matrixBlock.get_trace( inputParameters.dampingConstant )

outputFileName = inputParameters.path+str(inputParameters.frequency)+'_Tmatrix.out'
matrixBlock.print_block_to_file( outputFileName, inputParameters.frequency, inputParameters.dampingConstant )

